import logging
import numpy as np


def medfilt_custom(x, kernel_size=3):
    '''This median filter returns 'nan' whenever any value in the kernal width
    is 'nan' and the median otherwise'''
    T = x.shape[0]
    delta = kernel_size // 2

    x_med = np.zeros(x.shape)
    window = x[0:delta + 1]
    if np.any(np.isnan(window)):
        x_med[0] = np.nan
    else:
        x_med[0] = np.median(window)

    # print window
    for t in range(1, T):
        window = x[t - delta:t + delta + 1]
        # print window
        if np.any(np.isnan(window)):
            x_med[t] = np.nan
        else:
            x_med[t] = np.median(window)

    return x_med


def median_absolute_deviation(a, consistency_constant=1.4826):
    '''Calculate the median absolute deviation of a univariate dataset.

    Parameters
    ----------
    a : numpy.ndarray
        Sample data.
    consistency_constant : float
        Constant to make the MAD a consistent estimator of the population
        standard deviation (1.4826 for a normal distribution).

    Returns
    -------
    float
        Median absolute deviation of the data.
    '''
    return consistency_constant * np.nanmedian(np.abs(a - np.nanmedian(a)))


def post_process_cr(cr_params):
    """This will replace questionable values of the CR x and y position with
    'nan'.

        1)  threshold ellipse area by 99th percentile area distribution
        2)  median filter using custom median filter
        3)  remove deviations from discontinuous jumps

        The 'nan' values likely represent obscured CRs, secondary reflections,
        merges with the secondary reflection, or visual distortions due to
        the whisker or deformations of the eye

    Parameters
    ----------
    cr_params: numpy.ndarray
        (Nx5) array of pupil parameters [x, y, angle, axis1, axis2].
    """

    area = np.pi * (cr_params.T[3] / 2) * (cr_params.T[4] / 2)

    # compute a threshold on the area of the cr ellipse
    dev = median_absolute_deviation(area)
    if dev == 0:
        logging.warning("Median absolute deviation is 0,"
                        "falling back to standard deviation.")
        dev = np.nanstd(area)
    threshold = np.nanmedian(area) + 3 * dev

    x_center = cr_params.T[0]
    y_center = cr_params.T[1]

    # set x,y where area is over threshold to nan
    x_center[area > threshold] = np.nan
    y_center[area > threshold] = np.nan

    # median filter
    x_center_med = medfilt_custom(x_center, kernel_size=3)
    y_center_med = medfilt_custom(y_center, kernel_size=3)

    x_mask_finite = np.where(np.isfinite(x_center_med))[0]
    y_mask_finite = np.where(np.isfinite(y_center_med))[0]

    # if y increases discontinuously or x decreases discontinuously,
    # that is probably a CR secondary reflection
    mean_x = np.mean(x_center_med[x_mask_finite])
    mean_y = np.mean(y_center_med[y_mask_finite])

    std_x = np.std(x_center_med[x_mask_finite])
    std_y = np.std(y_center_med[y_mask_finite])

    # set these extreme values to nan
    x_center_med[np.abs(x_center_med - mean_x) > 3*std_x] = np.nan
    y_center_med[np.abs(y_center_med - mean_y) > 3*std_y] = np.nan

    either_nan_mask = np.isnan(x_center_med) | np.isnan(y_center_med)
    x_center_med[either_nan_mask] = np.nan
    y_center_med[either_nan_mask] = np.nan

    new_cr = np.vstack([x_center_med, y_center_med]).T

    bad_points_mask = either_nan_mask

    return new_cr, bad_points_mask


def post_process_areas(areas: np.ndarray, percent_thresh: int = 99):
    '''Filter pupil or eye area data by replacing outliers with nan

    Parameters
    ----------
    areas: np.ndarray
        (N x 1) Arra of ellipse areas for either eye or pupil
    percent_thresh: int
        Percentile to threshold at. Default is 99

    Returns
    -------
    numpy.ndarray
        Eye/pupil areas with outliers replaced with nan
    '''
    threshold = np.percentile(areas[np.isfinite(areas)], percent_thresh)
    outlier_indices = areas > threshold
    areas[outlier_indices] = np.nan
    return areas
